import re

from oscars_mgmnt.parser.es.es import ElasticsearchParser


class Sentinel2_ES_Parser(ElasticsearchParser):

    pattern_terrascope = re.compile(r"S2._\d{8}T\d{6}_[A-Z0-9]{5}_([A-Z]*)_V(\d{3})")
    terrascope_land_producttypes =  ["CCC", "CWC", "FAPAR", "FCOVER", "LAI", "NDVI", "TOC"]
    terrascope_water_producttypes = ["CHL", "RHOW", "SPM", "TUR"]
    wms_layer_mapping = {
        # Terrascope land
        "CCC": "CGS_S2_RADIOMETRY",
        "CWC": "CGS_S2_RADIOMETRY",
        "FAPAR": "CGS_S2_FAPAR",
        "FCOVER": "CGS_S2_FCOVER",
        "LAI": "CGS_S2_LAI",
        "NDVI": "CGS_S2_NDVI",
        "TOC": "CGS_S2_RADIOMETRY",
        # Terrascope water
        "CHL": "TERRASCOPE_S2_CHL_V1",
        "RHOW": "TERRASCOPE_S2_RHOW_V1",
        "SPM": "TERRASCOPE_S2_SPM_V1",
        "TUR": "TERRASCOPE_S2_TUR_V1"
    }

    def __init__(self, template=None, es_source_host=None, es_source_index=None):
        super().__init__('CGS_S2.template' if template is None else template, es_source_host, es_source_index)

    def init_dict(self):
        super().init_dict()
        self.dict_info.update(
            {
                "cloudcover": "properties.cloudcoverpercentage",
                "tileid": "properties.tileid",
                "processingbaseline": "properties.processingbaseline"
            }
        )

    def get_metadata(self, product_identifier):
        # replace product type with CGS, eg. S2A_20200323T030541_49QCE_TOC_V200 -> S2A_20200323T030541_49QCE_CGS_V200
        if self.pattern_terrascope.match(product_identifier):
            producttype = self.pattern_terrascope.search(product_identifier).group(1)
            if producttype in self.terrascope_land_producttypes:
                doc_id = product_identifier.replace(producttype, "CGS")
                return super().get_metadata(doc_id)
            elif producttype in self.terrascope_water_producttypes:
                doc_id = product_identifier.replace(producttype, "WATER")
                return super().get_metadata(doc_id)
        else:
            # fall back to original identifier
            return super().get_metadata(product_identifier)

    def specific_treatment(self, product_file_info, values):
        uid = super().specific_treatment(product_file_info, values)
        # fix geometry of L1C products
        # these may contain duplicate consecutive coordinates, which causes ES ingest to fail
        simplify = True if "_MSIL1C_" in uid else False
        self.fix_geometry(values, simplify=simplify)

        values['platformserialid'] = uid.split("_")[0]

        if self.pattern_terrascope.match(values['id']):
            producttype = self.pattern_terrascope.search(values['id']).group(1)
            values['producttype'] = producttype
            resolution_dir_regex = re.compile(r".*/(\d+)M")
            resolution_file_regex = re.compile(rf".*_{producttype}.*_(\d+)M_.*")

            if resolution_dir_regex.match(product_file_info.get_product_path()):
                # include resolution in product identifier, eg. FAPAR
                resolution = resolution_dir_regex.search(product_file_info.get_product_path()).group(1)
                values['resolution'] = int(resolution)
                values['id'] = f"{values['id'].rsplit('_', 1)[0]}_{resolution}M_{values['id'].rsplit('_', 1)[1]}"
            elif all(resolution_file_regex.match(file.name) for file in product_file_info.datas):
                resolutions = {resolution_file_regex.search(file.name).group(1) for file in product_file_info.datas}
                if len(resolutions) == 1:
                    # only one resolution, add resolution to product identifier, eg. CCC
                    resolution = resolutions.pop()
                    values['resolution'] = int(resolution)
                    values['id'] = f"{values['id'].rsplit('_', 1)[0]}_{resolution}M_{values['id'].rsplit('_', 1)[1]}"

            values['title'] = values['id']
            uid = values['id']
            values['parentidentifier'] = f"urn:eop:VITO:TERRASCOPE_S2_{values['producttype']}_{values['productVersion'][:2]}"
            if 'tileid' not in values:
                values['tileid'] = uid.split("_")[2]

            wms_layer = self.wms_layer_mapping[values['producttype']]
            self.add_wms_quicklook(wms_layer, product_file_info, values)
        else:
            # non-V200 product
            # change producttype from eg. S2MSI1C to L1C
            values['producttype'] = f"L{values['producttype'][-2:]}"
            values['parentidentifier'] = f"urn:eop:VITO:CGS_S2_{values['producttype']}"

            if "_OPER_" in uid and 'productVersion' in values:
                # for these kind of products a faulty processor version is extracted from the identifier
                values.pop('productVersion')

        return uid

    def fix_products(self, fileobject):
        """ helper method that moves files present in "datas" to "related".
        :param fileobject: the files associated to a product
        :type fileobject: a FileObject
        """

        # the ?: in front makes it non-capturing (so it doesn't introduce and additional group)
        shadow_pattern = re.compile(r'.*((?:SHADOWMASK_\d{2}M)|RAA_\d{2}M|VZA_\d{2}M|SZA_\d{2}M|WVP_\d{2}M|AOT_\d{2}M).*')
        cloud_pattern = re.compile(r'.*(CLOUDMASK_\d{2}M).*')
        scene_pattern = re.compile(r'.*(SCENECLASSIFICATION_\d{2}M).*')
        pixel_pattern = re.compile(r'.*(PIXELCLASSIFICATION_(\d{2}M)?).*')
        biopar_pattern = re.compile(r'.*((?:FAPAR|NDVI|FCOVER|LAI|CCC|CWC)_\d{2}M).*')
        water_param_pattern = re.compile(r'.*((?:CHL|SPM|TUR|CHL_SOURCE))_\d{2}M.*')
        radiometry_pattern = re.compile(r'.*((?:TOC-B\d\d_\d{2}M|TOC-B8A_20M)).*')
        water_rhow_pattern = re.compile(r'.*RHOW-((?:B\d\d|B8A))_\d{2}M.*')
        L1C_pattern = re.compile(r'.*_((?:TCI|B\d{2}|B8A))\..*')
        SAFE_pattern = re.compile(r'.*\.SAFE/(.*)')

        for info in list(fileobject.datas):

            if shadow_pattern.match(info.name):
                info.title = shadow_pattern.search(info.name).group(1)
                info.category = "QUALITY"
                fileobject.datas.remove(info)
                fileobject.related.append(info)
            elif cloud_pattern.match(info.name):
                info.title = cloud_pattern.search(info.name).group(1)
                info.category = "CLOUD"
                fileobject.datas.remove(info)
                fileobject.related.append(info)
            elif scene_pattern.match(info.name):
                info.title = scene_pattern.search(info.name).group(1)
                info.category = "QUALITY"
                fileobject.datas.remove(info)
                fileobject.related.append(info)
            elif pixel_pattern.match(info.name):
                info.title = pixel_pattern.search(info.name).group(1)
                info.category = "QUALITY"
                fileobject.datas.remove(info)
                fileobject.related.append(info)
            elif biopar_pattern.match(info.name):
                info.title = biopar_pattern.search(info.name).group(1)
            elif radiometry_pattern.match(info.name):
                info.title = radiometry_pattern.search(info.name).group(1)
            elif water_param_pattern.match(info.name):
                info.title = water_param_pattern.search(info.name).group(1)
            elif water_rhow_pattern.match(info.name):
                info.title = water_rhow_pattern.search(info.name).group(1)
            elif L1C_pattern.match(info.name):
                info.title = L1C_pattern.search(info.name).group(1)

        for alternate_info in list(fileobject.metadatas):
            # move XML files inside the SAFE folder to the related section
            if SAFE_pattern.match(alternate_info.name):
                fileobject.metadatas.remove(alternate_info)
                fileobject.related.append(alternate_info)

        for related_info in list(fileobject.related):
            if SAFE_pattern.match(related_info.name):
                related_info.title = SAFE_pattern.search(related_info.name).group(1)
                if related_info.type is None and related_info.name.endswith(".gml"):
                    related_info.type = "application/gml+xml"
                if related_info.category is None and "/QI_DATA/" in related_info.name:
                    related_info.category = "QUALITY"
